@@ -110,6 +110,10 @@ fn make_instruction_map() -> HashMap<String, (InstructionType, u8)> {
             entry(&format!("{}64", name), AluBinary, ebpf::BPF_ALU64 | opc);
         }
 
+        entry("sdiv", AluBinary, ebpf::BPF_ALU64 | ebpf::BPF_SDIV);
+        entry("sdiv64", AluBinary, ebpf::BPF_ALU64 | ebpf::BPF_SDIV);
+        entry("sdiv32", AluBinary, ebpf::BPF_ALU | ebpf::BPF_SDIV);
+
         // LoadAbs, LoadInd, LoadReg, StoreImm, and StoreReg.
         for &(suffix, size) in &mem_sizes {
             entry(
@@ -233,7 +237,9 @@ pub fn assemble<E: UserDefinedError, I: 'static + InstructionMeter>(
     }
 
     fn resolve_call(
+        config: &Config,
         bpf_functions: &mut BTreeMap<u32, (usize, String)>,
+        syscall_registry: &SyscallRegistry,
         labels: &HashMap<&str, usize>,
         label: &str,
         target_pc: Option<usize>,
@@ -245,7 +251,7 @@ pub fn assemble<E: UserDefinedError, I: 'static + InstructionMeter>(
                 .get(label)
                 .ok_or_else(|| format!("Label not found {}", label))?
         };
-        let hash = register_bpf_function(bpf_functions, target_pc, label, true)
+        let hash = register_bpf_function(config, bpf_functions, syscall_registry, target_pc, label)
             .map_err(|_| format!("Label hash collision {}", label))?;
         Ok(hash as i32 as i64)
     }
@@ -267,7 +273,14 @@ pub fn assemble<E: UserDefinedError, I: 'static + InstructionMeter>(
     }
     insn_ptr = 0;
     let mut bpf_functions = BTreeMap::new();
-    resolve_call(&mut bpf_functions, &labels, "entrypoint", None)?;
+    resolve_call(
+        &config,
+        &mut bpf_functions,
+        &syscall_registry,
+        &labels,
+        "entrypoint",
+        None,
+    )?;
     let mut instructions: Vec<Insn> = Vec::new();
     for statement in statements.iter() {
         if let Statement::Instruction { name, operands } = statement {
@@ -305,8 +318,14 @@ pub fn assemble<E: UserDefinedError, I: 'static + InstructionMeter>(
                         (CallImm, [Integer(imm)]) => {
                             let target_pc = (*imm + insn_ptr as i64 + 1) as usize;
                             let label = format!("function_{}", target_pc);
-                            let hash =
-                                resolve_call(&mut bpf_functions, &labels, &label, Some(target_pc))?;
+                            let hash = resolve_call(
+                                &config,
+                                &mut bpf_functions,
+                                &syscall_registry,
+                                &labels,
+                                &label,
+                                Some(target_pc),
+                            )?;
                             insn(opc, 0, 0, 0, hash as i32 as i64)
                         }
                         (CallReg, [Register(dst)]) => insn(opc, 0, 0, 0, *dst),
@@ -332,7 +351,14 @@ pub fn assemble<E: UserDefinedError, I: 'static + InstructionMeter>(
                             ebpf::hash_symbol_name(label.as_bytes()) as i32 as i64,
                         ),
                         (CallImm, [Label(label)]) => {
-                            let hash = resolve_call(&mut bpf_functions, &labels, label, None)?;
+                            let hash = resolve_call(
+                                &config,
+                                &mut bpf_functions,
+                                &syscall_registry,
+                                &labels,
+                                label,
+                                None,
+                            )?;
                             insn(opc, 0, 0, 0, hash as i32 as i64)
                         }
                         (Endian(size), [Register(dst)]) => insn(opc, *dst, 0, 0, size),
