@@ -27,7 +27,7 @@ use rand::{rngs::SmallRng, Rng, SeedableRng};
 use crate::{
     elf::Executable,
     vm::{Config, ProgramResult, InstructionMeter, Tracer, SYSCALL_CONTEXT_OBJECTS_OFFSET},
-    ebpf::{self, INSN_SIZE, FIRST_SCRATCH_REG, SCRATCH_REGS, STACK_REG, MM_STACK_START},
+    ebpf::{self, INSN_SIZE, FIRST_SCRATCH_REG, SCRATCH_REGS, FRAME_PTR_REG, MM_STACK_START, STACK_PTR_REG},
     error::{UserDefinedError, EbpfError},
     memory_region::{AccessType, MemoryMapping, MemoryRegion},
     user_error::UserError,
@@ -183,6 +183,8 @@ impl<E: UserDefinedError, I: InstructionMeter> JitProgram<E, I> {
 }
 
 // Special values for target_pc in struct Jump
+const TARGET_PC_LOCAL_ANCHOR: usize = std::usize::MAX - 100;
+const TARGET_PC_DIV_OVERFLOW: usize = std::usize::MAX - 33;
 const TARGET_PC_TRACE: usize = std::usize::MAX - 32;
 const TARGET_PC_SYSCALL: usize = std::usize::MAX - 31;
 const TARGET_PC_BPF_CALL_PROLOGUE: usize = std::usize::MAX - 30;
@@ -376,20 +378,32 @@ fn set_anchor(jit: &mut JitCompiler, target: usize) {
 enum EnvironmentStackSlot {
     /// The 6 CALLEE_SAVED_REGISTERS
     LastSavedRegister = 5,
-    /// REGISTER_MAP[STACK_REG]
-    BpfStackPtr = 6,
+    /// The current call depth.
+    ///
+    /// Incremented on calls and decremented on exits. It's used to enforce
+    /// config.max_call_depth and to know when to terminate execution.
+    CallDepth = 6,
+    /// BPF frame pointer (REGISTER_MAP[FRAME_PTR_REG]).
+    BpfFramePtr = 7,
+    /// The BPF stack pointer (r11). Only used when config.dynamic_stack_frames=true.
+    ///
+    /// The stack pointer isn't exposed as an actual register. Only sub and add
+    /// instructions (typically generated by the LLVM backend) are allowed to
+    /// access it. Its value is only stored in this slot and therefore the
+    /// register is not tracked in REGISTER_MAP.
+    BpfStackPtr = 8,
     /// Constant pointer to optional typed return value
-    OptRetValPtr = 7,
+    OptRetValPtr = 9,
     /// Last return value of instruction_meter.get_remaining()
-    PrevInsnMeter = 8,
+    PrevInsnMeter = 10,
     /// Constant pointer to instruction_meter
-    InsnMeterPtr = 9,
+    InsnMeterPtr = 11,
     /// CPU cycles accumulated by the stop watch
-    StopwatchNumerator = 10,
+    StopwatchNumerator = 12,
     /// Number of times the stop watch was used
-    StopwatchDenominator = 11,
+    StopwatchDenominator = 13,
     /// Bumper for size_of
-    SlotCount = 12,
+    SlotCount = 14,
 }
 
 fn slot_on_environment_stack(jit: &JitCompiler, slot: EnvironmentStackSlot) -> i32 {
@@ -578,7 +592,7 @@ fn emit_bpf_call<E: UserDefinedError>(jit: &mut JitCompiler, dst: Value) -> Resu
             // Move vm target_address into RAX
             X86Instruction::push(REGISTER_MAP[0], None).emit(jit)?;
             if reg != REGISTER_MAP[0] {
-                X86Instruction::mov(OperandSize::S64, reg, REGISTER_MAP[0]).emit(jit)?; 
+                X86Instruction::mov(OperandSize::S64, reg, REGISTER_MAP[0]).emit(jit)?;
             }
 
             emit_call(jit, TARGET_PC_BPF_CALL_REG)?;
@@ -602,7 +616,10 @@ fn emit_bpf_call<E: UserDefinedError>(jit: &mut JitCompiler, dst: Value) -> Resu
 
     emit_undo_profile_instruction_count(jit, 0)?;
 
-    X86Instruction::pop(REGISTER_MAP[STACK_REG]).emit(jit)?;
+    // Restore the previous frame pointer
+    X86Instruction::pop(REGISTER_MAP[FRAME_PTR_REG]).emit(jit)?;
+    let frame_ptr_access = X86IndirectAccess::Offset(slot_on_environment_stack(jit, EnvironmentStackSlot::BpfFramePtr));
+    X86Instruction::store(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], RBP, frame_ptr_access).emit(jit)?;
     for reg in REGISTER_MAP.iter().skip(FIRST_SCRATCH_REG).take(SCRATCH_REGS).rev() {
         X86Instruction::pop(*reg).emit(jit)?;
     }
@@ -801,14 +818,43 @@ fn emit_shift<E: UserDefinedError>(jit: &mut JitCompiler, size: OperandSize, opc
 fn emit_muldivmod<E: UserDefinedError>(jit: &mut JitCompiler, opc: u8, src: u8, dst: u8, imm: Option<i64>) -> Result<(), EbpfError<E>> {
     let mul = (opc & ebpf::BPF_ALU_OP_MASK) == (ebpf::MUL32_IMM & ebpf::BPF_ALU_OP_MASK);
     let div = (opc & ebpf::BPF_ALU_OP_MASK) == (ebpf::DIV32_IMM & ebpf::BPF_ALU_OP_MASK);
+    let sdiv = (opc & ebpf::BPF_ALU_OP_MASK) == (ebpf::SDIV32_IMM & ebpf::BPF_ALU_OP_MASK);
     let modrm = (opc & ebpf::BPF_ALU_OP_MASK) == (ebpf::MOD32_IMM & ebpf::BPF_ALU_OP_MASK);
     let size = if (opc & ebpf::BPF_CLS_MASK) == ebpf::BPF_ALU64 { OperandSize::S64 } else { OperandSize::S32 };
 
-    if (div || modrm) && imm.is_none() {
+    // subtracting offset_in_text_section from TARGET_PC_LOCAL_ANCHOR gives us a
+    // unique local anchor
+    let sdiv_anchor =  if sdiv { TARGET_PC_LOCAL_ANCHOR - jit.offset_in_text_section } else { 0 };
+
+    if (div || sdiv || modrm) && imm.is_none() {
         // Save pc
         X86Instruction::load_immediate(OperandSize::S64, R11, jit.pc as i64).emit(jit)?;
         X86Instruction::test(size, src, src, None).emit(jit)?; // src == 0
         emit_jcc(jit, 0x84, TARGET_PC_DIV_BY_ZERO)?;
+
+    }
+
+    // sdiv overflows with MIN / -1. If we have an immediate and it's not -1, we
+    // don't need any checks.
+    if sdiv && imm.unwrap_or(-1) == -1 {
+        if imm.is_none() {
+            // if src != -1, we can skip checking dst
+            X86Instruction::cmp_immediate(size, src, -1, None).emit(jit)?;
+            emit_jcc(jit, 0x85, sdiv_anchor)?;
+        }
+
+        // if dst != MIN, we're not going to overflow
+        X86Instruction::load_immediate(size, R11, if size == OperandSize::S64 { i64::MIN } else { i32::MIN as i64 }).emit(jit)?;
+        X86Instruction::cmp(size, dst, R11, None).emit(jit)?;
+        emit_jcc(jit, 0x85, sdiv_anchor)?;
+
+        // MIN / -1, raise EbpfError::DivideOverflow(pc)
+        X86Instruction::load_immediate(OperandSize::S64, R11, jit.pc as i64).emit(jit)?;
+        emit_jmp(jit, TARGET_PC_DIV_OVERFLOW)?;
+    }
+
+    if sdiv {
+        set_anchor(jit, sdiv_anchor);
     }
 
     if dst != RAX {
@@ -835,9 +881,17 @@ fn emit_muldivmod<E: UserDefinedError>(jit: &mut JitCompiler, opc: u8, src: u8,
     if div || modrm {
         // xor %edx,%edx
         emit_alu(jit, size, 0x31, RDX, RDX, 0, None)?;
+    } else if sdiv {
+        // cdq or cqo depending on operand size
+        X86Instruction {
+            size,
+            opcode: 0x99,
+            modrm: false,
+            ..X86Instruction::default()
+        }.emit(jit)?;
     }
 
-    emit_alu(jit, size, 0xf7, if mul { 4 } else { 6 }, R11, 0, None)?;
+    emit_alu(jit, size, 0xf7, if mul { 4 } else if sdiv { 7 } else { 6 }, R11, 0, None)?;
 
     if dst != RDX {
         if modrm {
@@ -846,7 +900,7 @@ fn emit_muldivmod<E: UserDefinedError>(jit: &mut JitCompiler, opc: u8, src: u8,
         X86Instruction::pop(RDX).emit(jit)?;
     }
     if dst != RAX {
-        if div || mul {
+        if div || sdiv || mul {
             X86Instruction::mov(OperandSize::S64, RAX, dst).emit(jit)?;
         }
         X86Instruction::pop(RAX).emit(jit)?;
@@ -931,21 +985,25 @@ impl std::fmt::Debug for JitCompiler {
 
 impl JitCompiler {
     // Arguments are unused on windows
-    fn new<E: UserDefinedError>(_program: &[u8], _config: &Config) -> Result<Self, EbpfError<E>> {
+    fn new<E: UserDefinedError>(program: &[u8], config: &Config) -> Result<Self, EbpfError<E>> {
         #[cfg(target_os = "windows")]
         {
+            let _ = program;
+            let _ = config;
             panic!("JIT not supported on windows");
         }
 
         #[cfg(not(target_arch = "x86_64"))]
         {
+            let _ = program;
+            let _ = config;
             panic!("JIT is only supported on x86_64");
         }
 
         // Scan through program to find actual number of instructions
         let mut pc = 0;
-        while (pc + 1) * ebpf::INSN_SIZE <= _program.len() {
-            let insn = ebpf::get_insn_unchecked(_program, pc);
+        while (pc + 1) * ebpf::INSN_SIZE <= program.len() {
+            let insn = ebpf::get_insn_unchecked(program, pc);
             pc += match insn.opc {
                 ebpf::LD_DW_IMM => 2,
                 _ => 1,
@@ -953,10 +1011,10 @@ impl JitCompiler {
         }
 
         let mut code_length_estimate = pc * 110 + 4096;
-        code_length_estimate += (code_length_estimate as f64 * _config.noop_instruction_ratio) as usize;
+        code_length_estimate += (code_length_estimate as f64 * config.noop_instruction_ratio) as usize;
         let mut diversification_rng = SmallRng::from_rng(rand::thread_rng()).unwrap();
-        let (environment_stack_key, program_argument_key) = 
-            if _config.encrypt_environment_registers {
+        let (environment_stack_key, program_argument_key) =
+            if config.encrypt_environment_registers {
                 (
                     diversification_rng.gen::<i32>() / 16, // -3 bits for 8 Byte alignment, and -1 bit to have encoding space for EnvironmentStackSlot::SlotCount
                     diversification_rng.gen::<i32>() / 2, // -1 bit to have encoding space for (SYSCALL_CONTEXT_OBJECTS_OFFSET + syscall.context_object_slot) * 8
@@ -972,7 +1030,7 @@ impl JitCompiler {
             last_instruction_meter_validation_pc: 0,
             program_vm_addr: 0,
             handler_anchors: HashMap::new(),
-            config: *_config,
+            config: *config,
             diversification_rng,
             stopwatch_is_active: false,
             environment_stack_key,
@@ -1015,11 +1073,22 @@ impl JitCompiler {
                 X86Instruction::load_immediate(OperandSize::S64, R11, 0).emit(self)?;
             }
 
-            let dst = REGISTER_MAP[insn.dst as usize];
+            let dst = if insn.dst == STACK_PTR_REG as u8 { u8::MAX } else { REGISTER_MAP[insn.dst as usize] };
             let src = REGISTER_MAP[insn.src as usize];
             let target_pc = (self.pc as isize + insn.off as isize + 1) as usize;
 
             match insn.opc {
+                _ if insn.dst == STACK_PTR_REG as u8 && self.config.dynamic_stack_frames => {
+                    let stack_ptr_access = X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfStackPtr));
+                    match insn.opc {
+                        ebpf::SUB64_IMM => emit_alu(self, OperandSize::S64, 0x81, 5, RBP, insn.imm, Some(stack_ptr_access))?,
+                        ebpf::ADD64_IMM => emit_alu(self, OperandSize::S64, 0x81, 0, RBP, insn.imm, Some(stack_ptr_access))?,
+                        _ => {
+                            #[cfg(debug_assertions)]
+                            unreachable!("unexpected insn on r11")
+                        }
+                    }
+                }
                 // BPF_LD class
                 ebpf::LD_ABS_B   => {
                     emit_address_translation(self, R11, Value::Constant64(ebpf::MM_INPUT_START.wrapping_add(insn.imm as u32 as u64) as i64, true), 1, AccessType::Load)?;
@@ -1137,9 +1206,9 @@ impl JitCompiler {
                     emit_alu(self, OperandSize::S32, 0x29, src, dst, 0, None)?;
                     X86Instruction::sign_extend_i32_to_i64(dst, dst).emit(self)?;
                 },
-                ebpf::MUL32_IMM | ebpf::DIV32_IMM | ebpf::MOD32_IMM  =>
+                ebpf::MUL32_IMM | ebpf::DIV32_IMM | ebpf::SDIV32_IMM | ebpf::MOD32_IMM  =>
                     emit_muldivmod(self, insn.opc, dst, dst, Some(insn.imm))?,
-                ebpf::MUL32_REG | ebpf::DIV32_REG | ebpf::MOD32_REG  =>
+                ebpf::MUL32_REG | ebpf::DIV32_REG | ebpf::SDIV32_REG | ebpf::MOD32_REG  =>
                     emit_muldivmod(self, insn.opc, src, dst, None)?,
                 ebpf::OR32_IMM   => emit_sanitized_alu(self, OperandSize::S32, 0x09, 1, dst, insn.imm)?,
                 ebpf::OR32_REG   => emit_alu(self, OperandSize::S32, 0x09, src, dst, 0, None)?,
@@ -1195,9 +1264,9 @@ impl JitCompiler {
                 ebpf::ADD64_REG  => emit_alu(self, OperandSize::S64, 0x01, src, dst, 0, None)?,
                 ebpf::SUB64_IMM  => emit_sanitized_alu(self, OperandSize::S64, 0x29, 5, dst, insn.imm)?,
                 ebpf::SUB64_REG  => emit_alu(self, OperandSize::S64, 0x29, src, dst, 0, None)?,
-                ebpf::MUL64_IMM | ebpf::DIV64_IMM | ebpf::MOD64_IMM  =>
+                ebpf::MUL64_IMM | ebpf::DIV64_IMM | ebpf::SDIV64_IMM | ebpf::MOD64_IMM  =>
                     emit_muldivmod(self, insn.opc, dst, dst, Some(insn.imm))?,
-                ebpf::MUL64_REG | ebpf::DIV64_REG | ebpf::MOD64_REG  =>
+                ebpf::MUL64_REG | ebpf::DIV64_REG | ebpf::SDIV64_REG | ebpf::MOD64_REG  =>
                     emit_muldivmod(self, insn.opc, src, dst, None)?,
                 ebpf::OR64_IMM   => emit_sanitized_alu(self, OperandSize::S64, 0x09, 1, dst, insn.imm)?,
                 ebpf::OR64_REG   => emit_alu(self, OperandSize::S64, 0x09, src, dst, 0, None)?,
@@ -1289,21 +1358,22 @@ impl JitCompiler {
                     emit_bpf_call(self, Value::Register(REGISTER_MAP[insn.imm as usize]))?;
                 },
                 ebpf::EXIT      => {
-                    let stack_ptr_access = X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfStackPtr));
-                    let stack_frame_size = self.config.stack_frame_size as i64 * if self.config.enable_stack_frame_gaps { 2 } else { 1 };
+                    let call_depth_access = X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::CallDepth));
+                    X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[FRAME_PTR_REG], call_depth_access).emit(self)?;
 
-                    // if((stack_ptr as u32) == self.config.stack_frame_size) goto exit;
-                    X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[STACK_REG], stack_ptr_access).emit(self)?;
-                    X86Instruction::cmp_immediate(OperandSize::S32, REGISTER_MAP[STACK_REG], self.config.stack_frame_size as i64, None).emit(self)?;
+                    // If CallDepth == 0, we've reached the exit instruction of the entry point
+                    X86Instruction::cmp_immediate(OperandSize::S32, REGISTER_MAP[FRAME_PTR_REG], 0, None).emit(self)?;
                     if self.config.enable_instruction_meter {
                         X86Instruction::load_immediate(OperandSize::S64, R11, self.pc as i64).emit(self)?;
                     }
+                    // we're done
                     emit_jcc(self, 0x84, TARGET_PC_EXIT)?;
 
-                    emit_alu(self, OperandSize::S64, 0x81, 5, REGISTER_MAP[STACK_REG], stack_frame_size, None)?; // stack_ptr -= stack_frame_size;
-                    X86Instruction::store(OperandSize::S64, REGISTER_MAP[STACK_REG], RBP, stack_ptr_access).emit(self)?;
+                    // else decrement and update CallDepth
+                    emit_alu(self, OperandSize::S64, 0x81, 5, REGISTER_MAP[FRAME_PTR_REG], 1, None)?;
+                    X86Instruction::store(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], RBP, call_depth_access).emit(self)?;
 
-                    // else return;
+                    // and return
                     emit_validate_and_profile_instruction_count(self, false, Some(0))?;
                     X86Instruction::return_near().emit(self)?;
                 },
@@ -1397,15 +1467,31 @@ impl JitCompiler {
         for (i, reg) in REGISTER_MAP.iter().skip(FIRST_SCRATCH_REG).take(SCRATCH_REGS).enumerate() {
             X86Instruction::store(OperandSize::S64, *reg, RSP, X86IndirectAccess::OffsetIndexShift(8 * (SCRATCH_REGS - i + 1) as i32, RSP, 0)).emit(self)?; // Push SCRATCH_REG
         }
-        X86Instruction::store(OperandSize::S64, REGISTER_MAP[STACK_REG], RSP, X86IndirectAccess::OffsetIndexShift(8, RSP, 0)).emit(self)?; // Push REGISTER_MAP[STACK_REG]
+        // Push the caller's frame pointer. The code to restore it is emitted at the end of emit_bpf_call().
+        X86Instruction::store(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], RSP, X86IndirectAccess::OffsetIndexShift(8, RSP, 0)).emit(self)?;
         X86Instruction::xchg(OperandSize::S64, R11, RSP, Some(X86IndirectAccess::OffsetIndexShift(0, RSP, 0))).emit(self)?; // Push return address and restore original R11
-        let stack_ptr_access = X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfStackPtr));
-        let stack_frame_size = self.config.stack_frame_size as i64 * if self.config.enable_stack_frame_gaps { 2 } else { 1 };
-        emit_alu(self, OperandSize::S64, 0x81, 0, RBP, stack_frame_size, Some(stack_ptr_access))?; // stack_ptr += stack_frame_size;
-        X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[STACK_REG], stack_ptr_access).emit(self)?; // Load BpfStackPtr
-        // if((stack_ptr as u32) >= self.config.stack_frame_size + self.config.max_call_depth * stack_frame_size) throw EbpfError::CallDepthExeeded;
-        X86Instruction::cmp_immediate(OperandSize::S32, REGISTER_MAP[STACK_REG], self.config.stack_frame_size as i64 + (self.config.max_call_depth as i64 * stack_frame_size), None).emit(self)?;
+
+        // Increase CallDepth
+        let call_depth_access = X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::CallDepth));
+        emit_alu(self, OperandSize::S64, 0x81, 0, RBP, 1, Some(call_depth_access))?;
+        X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[FRAME_PTR_REG], call_depth_access).emit(self)?;
+        // If CallDepth == self.config.max_call_depth, stop and return CallDepthExceeded
+        X86Instruction::cmp_immediate(OperandSize::S32, REGISTER_MAP[FRAME_PTR_REG], self.config.max_call_depth as i64, None).emit(self)?;
         emit_jcc(self, 0x83, TARGET_PC_CALL_DEPTH_EXCEEDED)?;
+
+        // Setup the frame pointer for the new frame. What we do depends on whether we're using dynamic or fixed frames.
+        let frame_ptr_access = X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfFramePtr));
+        if self.config.dynamic_stack_frames {
+            // When dynamic frames are on, the next frame starts at the end of the current frame
+            let stack_ptr_access = X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfStackPtr));
+            X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[FRAME_PTR_REG], stack_ptr_access).emit(self)?;
+            X86Instruction::store(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], RBP, frame_ptr_access).emit(self)?;
+        } else {
+            // With fixed frames we start the new frame at the next fixed offset
+            let stack_frame_size = self.config.stack_frame_size as i64 * if self.config.enable_stack_frame_gaps { 2 } else { 1 };
+            emit_alu(self, OperandSize::S64, 0x81, 0, RBP, stack_frame_size, Some(frame_ptr_access))?; // frame_ptr += stack_frame_size;
+            X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[FRAME_PTR_REG], frame_ptr_access).emit(self)?; // Load BpfFramePtr
+        }
         X86Instruction::return_near().emit(self)?;
 
         // Routine for emit_bpf_call(Value::Register())
@@ -1415,16 +1501,16 @@ impl JitCompiler {
         // Upper bound check
         // if(RAX >= self.program_vm_addr + number_of_instructions * INSN_SIZE) throw CALL_OUTSIDE_TEXT_SEGMENT;
         let number_of_instructions = self.result.pc_section.len() - 1;
-        X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[STACK_REG], self.program_vm_addr as i64 + (number_of_instructions * INSN_SIZE) as i64).emit(self)?;
-        X86Instruction::cmp(OperandSize::S64, REGISTER_MAP[STACK_REG], REGISTER_MAP[0], None).emit(self)?;
+        X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], self.program_vm_addr as i64 + (number_of_instructions * INSN_SIZE) as i64).emit(self)?;
+        X86Instruction::cmp(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], REGISTER_MAP[0], None).emit(self)?;
         emit_jcc(self, 0x83, TARGET_PC_CALL_OUTSIDE_TEXT_SEGMENT)?;
         // Lower bound check
         // if(RAX < self.program_vm_addr) throw CALL_OUTSIDE_TEXT_SEGMENT;
-        X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[STACK_REG], self.program_vm_addr as i64).emit(self)?;
-        X86Instruction::cmp(OperandSize::S64, REGISTER_MAP[STACK_REG], REGISTER_MAP[0], None).emit(self)?;
+        X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], self.program_vm_addr as i64).emit(self)?;
+        X86Instruction::cmp(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], REGISTER_MAP[0], None).emit(self)?;
         emit_jcc(self, 0x82, TARGET_PC_CALL_OUTSIDE_TEXT_SEGMENT)?;
         // Calculate offset relative to instruction_addresses
-        emit_alu(self, OperandSize::S64, 0x29, REGISTER_MAP[STACK_REG], REGISTER_MAP[0], 0, None)?; // RAX -= self.program_vm_addr;
+        emit_alu(self, OperandSize::S64, 0x29, REGISTER_MAP[FRAME_PTR_REG], REGISTER_MAP[0], 0, None)?; // RAX -= self.program_vm_addr;
         // Calculate the target_pc (dst / INSN_SIZE) to update the instruction_meter
         let shift_amount = INSN_SIZE.trailing_zeros();
         debug_assert_eq!(INSN_SIZE, 1 << shift_amount);
@@ -1434,11 +1520,11 @@ impl JitCompiler {
         X86Instruction::store(OperandSize::S64, R11, RSP, X86IndirectAccess::OffsetIndexShift(-8, RSP, 0)).emit(self)?; // RSP[-8] = R11;
         // Load host target_address from self.result.pc_section
         debug_assert_eq!(INSN_SIZE, 8); // Because the instruction size is also the slot size we do not need to shift the offset
-        X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[STACK_REG], self.result.pc_section.as_ptr() as i64).emit(self)?;
-        emit_alu(self, OperandSize::S64, 0x01, REGISTER_MAP[STACK_REG], REGISTER_MAP[0], 0, None)?; // RAX += self.result.pc_section;
+        X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], self.result.pc_section.as_ptr() as i64).emit(self)?;
+        emit_alu(self, OperandSize::S64, 0x01, REGISTER_MAP[FRAME_PTR_REG], REGISTER_MAP[0], 0, None)?; // RAX += self.result.pc_section;
         X86Instruction::load(OperandSize::S64, REGISTER_MAP[0], REGISTER_MAP[0], X86IndirectAccess::Offset(0)).emit(self)?; // RAX = self.result.pc_section[RAX / 8];
-        // Load BpfStackPtr
-        X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[STACK_REG], X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfStackPtr))).emit(self)?;
+        // Load the frame pointer again since we've clobbered REGISTER_MAP[FRAME_PTR_REG]
+        X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[FRAME_PTR_REG], X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfFramePtr))).emit(self)?;
         X86Instruction::return_near().emit(self)?;
 
         // Translates a host pc back to a BPF pc by linear search of the pc_section table
@@ -1473,7 +1559,7 @@ impl JitCompiler {
             X86Instruction::push(R11, None).emit(self)?;
             X86Instruction::push(RAX, None).emit(self)?;
             X86Instruction::push(RCX, None).emit(self)?;
-            let stack_offset = if self.config.enable_stack_frame_gaps {
+            let stack_offset = if !self.config.dynamic_stack_frames && self.config.enable_stack_frame_gaps {
                 X86Instruction::push(RDX, None).emit(self)?;
                 24
             } else {
@@ -1490,9 +1576,11 @@ impl JitCompiler {
                 X86Instruction::cmp_immediate(OperandSize::S8, RAX, 0, Some(X86IndirectAccess::Offset(MemoryRegion::IS_WRITABLE_OFFSET))).emit(self)?; // region.is_writable == 0
                 emit_jcc(self, 0x84, TARGET_PC_MEMORY_ACCESS_VIOLATION + target_offset)?;
             }
-            X86Instruction::load_immediate(OperandSize::S64, RCX, (1i64 << ebpf::VIRTUAL_ADDRESS_BITS) - 1).emit(self)?; // RCX = (1 << ebpf::VIRTUAL_ADDRESS_BITS) - 1;
-            emit_alu(self, OperandSize::S64, 0x21, RCX, R11, 0, None)?; // R11 &= (1 << ebpf::VIRTUAL_ADDRESS_BITS) - 1;
-            if self.config.enable_stack_frame_gaps {
+            X86Instruction::load(OperandSize::S64, RAX, RCX, X86IndirectAccess::Offset(MemoryRegion::VM_ADDR_OFFSET)).emit(self)?; // RCX = region.vm_addr
+            X86Instruction::cmp(OperandSize::S64, RCX, R11, None).emit(self)?; // vm_addr < region.vm_addr
+            emit_jcc(self, 0x82, TARGET_PC_MEMORY_ACCESS_VIOLATION + target_offset)?;
+            emit_alu(self, OperandSize::S64, 0x29, RCX, R11, 0, None)?; // vm_addr -= region.vm_addr
+            if !self.config.dynamic_stack_frames && self.config.enable_stack_frame_gaps {
                 X86Instruction::load(OperandSize::S8, RAX, RCX, X86IndirectAccess::Offset(MemoryRegion::VM_GAP_SHIFT_OFFSET)).emit(self)?; // RCX = region.vm_gap_shift;
                 X86Instruction::mov(OperandSize::S64, R11, RDX).emit(self)?; // RDX = R11;
                 emit_alu(self, OperandSize::S64, 0xd3, 5, RDX, 0, None)?; // RDX = R11 >> region.vm_gap_shift;
@@ -1511,7 +1599,7 @@ impl JitCompiler {
             X86Instruction::cmp(OperandSize::S64, RCX, RAX, Some(X86IndirectAccess::Offset(MemoryRegion::LEN_OFFSET))).emit(self)?; // region.len < R11 + len
             emit_jcc(self, 0x82, TARGET_PC_MEMORY_ACCESS_VIOLATION + target_offset)?;
             emit_alu(self, OperandSize::S64, 0x03, R11, RAX, 0, Some(X86IndirectAccess::Offset(MemoryRegion::HOST_ADDR_OFFSET)))?; // R11 += region.host_addr;
-            if self.config.enable_stack_frame_gaps {
+            if !self.config.dynamic_stack_frames && self.config.enable_stack_frame_gaps {
                 X86Instruction::pop(RDX).emit(self)?;
             }
             X86Instruction::pop(RCX).emit(self)?;
@@ -1562,6 +1650,11 @@ impl JitCompiler {
         emit_set_exception_kind::<E>(self, EbpfError::DivideByZero(0))?;
         emit_jmp(self, TARGET_PC_EXCEPTION_AT)?;
 
+        // Handler for EbpfError::DivideOverflow
+        set_anchor(self, TARGET_PC_DIV_OVERFLOW);
+        emit_set_exception_kind::<E>(self, EbpfError::DivideOverflow(0))?;
+        emit_jmp(self, TARGET_PC_EXCEPTION_AT)?;
+
         // Handler for EbpfError::UnsupportedInstruction
         set_anchor(self, TARGET_PC_CALLX_UNSUPPORTED_INSTRUCTION);
         // Load BPF target pc from stack (which was saved in TARGET_PC_BPF_CALL_REG)
@@ -1599,9 +1692,27 @@ impl JitCompiler {
             X86Instruction::push(*reg, None).emit(self)?;
         }
 
-        // Initialize and save BPF stack pointer
-        X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[STACK_REG], MM_STACK_START as i64 + self.config.stack_frame_size as i64).emit(self)?;
-        X86Instruction::push(REGISTER_MAP[STACK_REG], None).emit(self)?;
+        // Initialize CallDepth to 0
+        X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], 0).emit(self)?;
+        X86Instruction::push(REGISTER_MAP[FRAME_PTR_REG], None).emit(self)?;
+
+        // Initialize the BPF frame and stack pointers (BpfFramePtr and BpfStackPtr)
+        if self.config.dynamic_stack_frames {
+            // The stack is fully descending from MM_STACK_START + stack_size to MM_STACK_START
+            X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], MM_STACK_START as i64 + self.config.stack_size() as i64).emit(self)?;
+            // Push BpfFramePtr
+            X86Instruction::push(REGISTER_MAP[FRAME_PTR_REG], None).emit(self)?;
+            // Push BpfStackPtr
+            X86Instruction::push(REGISTER_MAP[FRAME_PTR_REG], None).emit(self)?;
+        } else {
+            // The frames are ascending from MM_STACK_START to MM_STACK_START + stack_size. The stack within the frames is descending.
+            X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[FRAME_PTR_REG], MM_STACK_START as i64 + self.config.stack_frame_size as i64).emit(self)?;
+            // Push BpfFramePtr
+            X86Instruction::push(REGISTER_MAP[FRAME_PTR_REG], None).emit(self)?;
+            // When using static frames BpfStackPtr is not used
+            X86Instruction::load_immediate(OperandSize::S64, RBP, 0).emit(self)?;
+            X86Instruction::push(RBP, None).emit(self)?;
+        }
 
         // Save pointer to optional typed return value
         X86Instruction::push(ARGUMENT_REGISTERS[0], None).emit(self)?;
@@ -1616,20 +1727,20 @@ impl JitCompiler {
         X86Instruction::push(ARGUMENT_REGISTERS[3], None).emit(self)?;
 
         // Initialize stop watch
-        emit_alu(self, OperandSize::S64, 0x31, R11, R11, 0, None)?; // R11 ^= R11; 
+        emit_alu(self, OperandSize::S64, 0x31, R11, R11, 0, None)?; // R11 ^= R11;
         X86Instruction::push(R11, None).emit(self)?;
         X86Instruction::push(R11, None).emit(self)?;
 
         // Initialize frame pointer
         X86Instruction::mov(OperandSize::S64, RSP, RBP).emit(self)?;
         emit_alu(self, OperandSize::S64, 0x81, 0, RBP, 8 * (EnvironmentStackSlot::SlotCount as i64 - 1 + self.environment_stack_key as i64), None)?;
-        
+
         // Save JitProgramArgument
         X86Instruction::lea(OperandSize::S64, ARGUMENT_REGISTERS[2], R10, Some(X86IndirectAccess::Offset(-self.program_argument_key))).emit(self)?;
 
         // Zero BPF registers
         for reg in REGISTER_MAP.iter() {
-            if *reg != REGISTER_MAP[1] && *reg != REGISTER_MAP[STACK_REG] {
+            if *reg != REGISTER_MAP[1] && *reg != REGISTER_MAP[FRAME_PTR_REG] {
                 X86Instruction::load_immediate(OperandSize::S64, *reg, 0).emit(self)?;
             }
         }
