@@ -32,56 +32,12 @@ use crate::{
 use libc::c_char;
 use std::{slice::from_raw_parts, str::from_utf8, u64};
 
+/// Test syscall context
+pub type BpfSyscallContext = u64;
+
 /// Return type of syscalls
 pub type Result = std::result::Result<u64, EbpfError<UserError>>;
 
-// syscalls associated to kernel syscalls
-// See also linux/include/uapi/linux/bpf.h in Linux kernel sources.
-
-// bpf_ktime_getns()
-
-/// Index of syscall `bpf_ktime_getns()`, equivalent to `bpf_time_getns()`, in Linux kernel, see
-/// <https://git.kernel.org/cgit/linux/kernel/git/torvalds/linux.git/tree/include/uapi/linux/bpf.h>.
-pub const BPF_KTIME_GETNS_IDX: u32 = 5;
-
-/// Get monotonic time (since boot time) in nanoseconds. All arguments are unused.
-///
-/// # Examples
-///
-/// ```
-/// use solana_rbpf::syscalls::{BpfTimeGetNs, Result};
-/// use solana_rbpf::memory_region::{MemoryRegion, MemoryMapping};
-/// use solana_rbpf::vm::{Config, SyscallObject};
-/// use solana_rbpf::user_error::UserError;
-///
-/// let mut result: Result = Ok(0);
-/// let config = Config::default();
-/// let memory_mapping = MemoryMapping::new::<UserError>(vec![], &config).unwrap();
-/// BpfTimeGetNs::call(&mut BpfTimeGetNs {}, 0, 0, 0, 0, 0, &memory_mapping, &mut result);
-/// let t = result.unwrap();
-/// let d =  t / 10u64.pow(9)  / 60   / 60  / 24;
-/// let h = (t / 10u64.pow(9)  / 60   / 60) % 24;
-/// let m = (t / 10u64.pow(9)  / 60 ) % 60;
-/// let s = (t / 10u64.pow(9)) % 60;
-/// let ns = t % 10u64.pow(9);
-/// println!("Uptime: {:#x} == {} days {}:{}:{}, {} ns", t, d, h, m, s, ns);
-/// ```
-pub struct BpfTimeGetNs {}
-impl SyscallObject<UserError> for BpfTimeGetNs {
-    fn call(
-        &mut self,
-        _arg1: u64,
-        _arg2: u64,
-        _arg3: u64,
-        _arg4: u64,
-        _arg5: u64,
-        _memory_mapping: &MemoryMapping,
-        result: &mut Result,
-    ) {
-        *result = Result::Ok(time::precise_time_ns());
-    }
-}
-
 // bpf_trace_printk()
 
 /// Index of syscall `bpf_trace_printk()`, equivalent to `bpf_trace_printf`, in Linux kernel, see
@@ -130,6 +86,12 @@ pub const BPF_TRACE_PRINTK_IDX: u32 = 6;
 /// This would equally print the three numbers in `/sys/kernel/debug/tracing` file each time the
 /// program is run.
 pub struct BpfTracePrintf {}
+impl BpfTracePrintf {
+    /// new
+    pub fn init<C, E>(_unused: C) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self {})
+    }
+}
 impl SyscallObject<UserError> for BpfTracePrintf {
     fn call(
         &mut self,
@@ -178,6 +140,12 @@ impl SyscallObject<UserError> for BpfTracePrintf {
 /// assert_eq!(result.unwrap(), 0x1122334455);
 /// ```
 pub struct BpfGatherBytes {}
+impl BpfGatherBytes {
+    /// new
+    pub fn init<C, E>(_unused: C) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self {})
+    }
+}
 impl SyscallObject<UserError> for BpfGatherBytes {
     fn call(
         &mut self,
@@ -223,6 +191,12 @@ impl SyscallObject<UserError> for BpfGatherBytes {
 /// assert_eq!(val, vec![0x00, 0x00, 0x00, 0x00, 0x00, 0x11, 0x22, 0x33]);
 /// ```
 pub struct BpfMemFrob {}
+impl BpfMemFrob {
+    /// new
+    pub fn init<C, E>(_unused: C) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self {})
+    }
+}
 impl SyscallObject<UserError> for BpfMemFrob {
     fn call(
         &mut self,
@@ -263,6 +237,12 @@ impl SyscallObject<UserError> for BpfMemFrob {
 /// assert_eq!(result.unwrap(), 3);
 /// ```
 pub struct BpfSqrtI {}
+impl BpfSqrtI {
+    /// new
+    pub fn init<C, E>(_unused: C) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self {})
+    }
+}
 impl SyscallObject<UserError> for BpfSqrtI {
     fn call(
         &mut self,
@@ -304,6 +284,12 @@ impl SyscallObject<UserError> for BpfSqrtI {
 /// assert!(result.unwrap() != 0);
 /// ```
 pub struct BpfStrCmp {}
+impl BpfStrCmp {
+    /// new
+    pub fn init<C, E>(_unused: C) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self {})
+    }
+}
 impl SyscallObject<UserError> for BpfStrCmp {
     fn call(
         &mut self,
@@ -353,17 +339,12 @@ impl SyscallObject<UserError> for BpfStrCmp {
 /// ```
 /// extern crate libc;
 /// extern crate solana_rbpf;
-/// extern crate time;
 ///
 /// use solana_rbpf::syscalls::{BpfRand, Result};
 /// use solana_rbpf::memory_region::{MemoryRegion, MemoryMapping};
 /// use solana_rbpf::vm::{Config, SyscallObject};
 /// use solana_rbpf::user_error::UserError;
-///
-/// unsafe {
-///     libc::srand(time::precise_time_ns() as u32)
-/// }
-///
+//////
 /// let mut result: Result = Ok(0);
 /// let config = Config::default();
 /// let memory_mapping = MemoryMapping::new::<UserError>(vec![], &config).unwrap();
@@ -372,6 +353,12 @@ impl SyscallObject<UserError> for BpfStrCmp {
 /// assert!(3 <= n && n <= 6);
 /// ```
 pub struct BpfRand {}
+impl BpfRand {
+    /// new
+    pub fn init<C, E>(_unused: C) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self {})
+    }
+}
 impl SyscallObject<UserError> for BpfRand {
     fn call(
         &mut self,
@@ -393,6 +380,12 @@ impl SyscallObject<UserError> for BpfRand {
 
 /// Prints a NULL-terminated UTF-8 string.
 pub struct BpfSyscallString {}
+impl BpfSyscallString {
+    /// new
+    pub fn init<C, E>(_unused: C) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self {})
+    }
+}
 impl SyscallObject<UserError> for BpfSyscallString {
     fn call(
         &mut self,
@@ -423,6 +416,12 @@ impl SyscallObject<UserError> for BpfSyscallString {
 
 /// Prints the five arguments formated as u64 in decimal.
 pub struct BpfSyscallU64 {}
+impl BpfSyscallU64 {
+    /// new
+    pub fn init<C, E>(_unused: C) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self {})
+    }
+}
 impl SyscallObject<UserError> for BpfSyscallU64 {
     fn call(
         &mut self,
@@ -442,10 +441,16 @@ impl SyscallObject<UserError> for BpfSyscallU64 {
     }
 }
 
-/// Excample of a syscall with internal state.
+/// Example of a syscall with internal state.
 pub struct SyscallWithContext {
     /// Mutable state
-    pub context: u64,
+    pub context: BpfSyscallContext,
+}
+impl SyscallWithContext {
+    /// new
+    pub fn init<C, E>(context: BpfSyscallContext) -> Box<dyn SyscallObject<UserError>> {
+        Box::new(Self { context })
+    }
 }
 impl SyscallObject<UserError> for SyscallWithContext {
     fn call(
