@@ -12,6 +12,7 @@
 //! Virtual machine and JIT compiler for eBPF programs.
 
 use crate::disassembler::disassemble_instruction;
+use crate::ebpf::STACK_PTR_REG;
 use crate::static_analysis::Analysis;
 use crate::{
     call_frames::CallFrames,
@@ -60,6 +61,9 @@ macro_rules! question_mark {
     }};
 }
 
+/// Syscall initialization function
+pub type SyscallInit<'a, C, E> = fn(C) -> Box<(dyn SyscallObject<E> + 'a)>;
+
 /// Syscall function without context
 pub type SyscallFunction<E, O> =
     fn(O, u64, u64, u64, u64, u64, &MemoryMapping, &mut ProgramResult<E>);
@@ -83,6 +87,8 @@ pub trait SyscallObject<E: UserDefinedError> {
 /// Syscall function and binding slot for a context object
 #[derive(Debug, PartialEq)]
 pub struct Syscall {
+    /// Syscall init
+    pub init: u64,
     /// Call the syscall function
     pub function: u64,
     /// Slot of context object
@@ -122,11 +128,13 @@ pub struct SyscallRegistry {
 
 impl SyscallRegistry {
     /// Register a syscall function by its symbol hash
-    pub fn register_syscall_by_hash<E: UserDefinedError, O: SyscallObject<E>>(
+    pub fn register_syscall_by_hash<'a, C, E: UserDefinedError, O: SyscallObject<E>>(
         &mut self,
         hash: u32,
+        init: SyscallInit<'a, C, E>,
         function: SyscallFunction<E, &mut O>,
     ) -> Result<(), EbpfError<E>> {
+        let init = init as *const u8 as u64;
         let function = function as *const u8 as u64;
         let context_object_slot = self.entries.len();
         if self
@@ -134,6 +142,7 @@ impl SyscallRegistry {
             .insert(
                 hash,
                 Syscall {
+                    init,
                     function,
                     context_object_slot,
                 },
@@ -151,12 +160,13 @@ impl SyscallRegistry {
     }
 
     /// Register a syscall function by its symbol name
-    pub fn register_syscall_by_name<E: UserDefinedError, O: SyscallObject<E>>(
+    pub fn register_syscall_by_name<'a, C, E: UserDefinedError, O: SyscallObject<E>>(
         &mut self,
         name: &[u8],
+        init: SyscallInit<'a, C, E>,
         function: SyscallFunction<E, &mut O>,
     ) -> Result<(), EbpfError<E>> {
-        self.register_syscall_by_hash(ebpf::hash_symbol_name(name), function)
+        self.register_syscall_by_hash::<C, E, O>(ebpf::hash_symbol_name(name), init, function)
     }
 
     /// Get a symbol's function pointer and context object slot
@@ -211,6 +221,23 @@ pub struct Config {
     pub encrypt_environment_registers: bool,
     /// Disable ldabs* and ldind* instructions
     pub disable_deprecated_load_instructions: bool,
+    /// Throw ElfError::SymbolHashCollision when a BPF function collides with a registered syscall
+    pub syscall_bpf_function_hash_collision: bool,
+    /// Have the verifier reject "callx r10"
+    pub reject_callx_r10: bool,
+    /// Use dynamic stack frame sizes
+    pub dynamic_stack_frames: bool,
+    /// Enable native signed division
+    pub enable_sdiv: bool,
+    /// Avoid copying read only sections when possible
+    pub optimize_rodata: bool,
+}
+
+impl Config {
+    /// Returns the size of the stack memory region
+    pub fn stack_size(&self) -> usize {
+        self.stack_frame_size * self.max_call_depth
+    }
 }
 
 impl Default for Config {
@@ -225,10 +252,15 @@ impl Default for Config {
             enable_symbol_and_section_labels: false,
             disable_unresolved_symbols_at_runtime: true,
             reject_broken_elfs: false,
-            disable_deprecated_load_instructions: true,
             noop_instruction_ratio: 1.0 / 256.0,
             sanitize_user_provided_values: true,
             encrypt_environment_registers: true,
+            disable_deprecated_load_instructions: true,
+            syscall_bpf_function_hash_collision: true,
+            reject_callx_r10: true,
+            dynamic_stack_frames: false,
+            enable_sdiv: true,
+            optimize_rodata: true,
         }
     }
 }
@@ -247,7 +279,7 @@ impl<E: UserDefinedError, I: 'static + InstructionMeter> Executable<E, I> {
     ) -> Result<Pin<Box<Self>>, EbpfError<E>> {
         let executable = Executable::load(config, elf_bytes, syscall_registry)?;
         if let Some(verifier) = verifier {
-            verifier(executable.get_text_bytes().1, &config)?;
+            verifier(executable.get_text_bytes().1, executable.get_config())?;
         }
         Ok(Pin::new(Box::new(executable)))
     }
@@ -441,9 +473,11 @@ macro_rules! translate_memory_access {
 /// ];
 ///
 /// // Instantiate a VM.
+/// let config = Config::default();
 /// let mut bpf_functions = std::collections::BTreeMap::new();
-/// register_bpf_function(&mut bpf_functions, 0, "entrypoint", false).unwrap();
-/// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
+/// let syscall_registry = SyscallRegistry::default();
+/// register_bpf_function(&config, &mut bpf_functions, &syscall_registry, 0, "entrypoint").unwrap();
+/// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, config, syscall_registry, bpf_functions).unwrap();
 /// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(&executable, &mut [], mem).unwrap();
 ///
 /// // Provide a reference to the packet data.
@@ -477,9 +511,11 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// ];
     ///
     /// // Instantiate a VM.
+    /// let config = Config::default();
     /// let mut bpf_functions = std::collections::BTreeMap::new();
-    /// register_bpf_function(&mut bpf_functions, 0, "entrypoint", false).unwrap();
-    /// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
+    /// let syscall_registry = SyscallRegistry::default();
+    /// register_bpf_function(&config, &mut bpf_functions, &syscall_registry, 0, "entrypoint").unwrap();
+    /// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, config, syscall_registry, bpf_functions).unwrap();
     /// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(&executable, &mut [], &mut []).unwrap();
     /// ```
     pub fn new(
@@ -488,11 +524,10 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
         input_region: &mut [u8],
     ) -> Result<EbpfVm<'a, E, I>, EbpfError<E>> {
         let config = executable.get_config();
-        let ro_region = executable.get_ro_section();
         let stack = CallFrames::new(config);
         let regions: Vec<MemoryRegion> = vec![
             MemoryRegion::new_from_slice(&[], 0, 0, false),
-            MemoryRegion::new_from_slice(ro_region, ebpf::MM_PROGRAM_START, 0, false),
+            executable.get_ro_region(),
             stack.get_memory_region(),
             MemoryRegion::new_from_slice(heap_region, ebpf::MM_HEAP_START, 0, true),
             MemoryRegion::new_from_slice(input_region, ebpf::MM_INPUT_START, 0, true),
@@ -539,7 +574,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
         &self.tracer
     }
 
-    /// Bind a context object instance to a previously registered syscall
+    /// Bind a context objects instance to a previously registered syscalls
     ///
     /// # Examples
     ///
@@ -565,44 +600,55 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// // On running the program this syscall will print the content of registers r3, r4 and r5 to
     /// // standard output.
     /// let mut syscall_registry = SyscallRegistry::default();
-    /// syscall_registry.register_syscall_by_hash(6, BpfTracePrintf::call).unwrap();
+    /// syscall_registry.register_syscall_by_hash(6, BpfTracePrintf::init::<u64, UserError>, BpfTracePrintf::call).unwrap();
     /// // Instantiate an Executable and VM
+    /// let config = Config::default();
     /// let mut bpf_functions = std::collections::BTreeMap::new();
-    /// register_bpf_function(&mut bpf_functions, 0, "entrypoint", false).unwrap();
-    /// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, Config::default(), syscall_registry, bpf_functions).unwrap();
+    /// register_bpf_function(&config, &mut bpf_functions, &syscall_registry, 0, "entrypoint").unwrap();
+    /// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, config, syscall_registry, bpf_functions).unwrap();
     /// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(&executable, &mut [], &mut []).unwrap();
     /// // Bind a context object instance to the previously registered syscall
-    /// vm.bind_syscall_context_object(Box::new(BpfTracePrintf {}), None);
+    /// vm.bind_syscall_context_objects(0, None);
     /// ```
-    pub fn bind_syscall_context_object(
+    pub fn bind_syscall_context_objects<C: Clone>(
         &mut self,
-        syscall_context_object: Box<dyn SyscallObject<E> + 'a>,
+        syscall_context: C,
         hash: Option<u32>,
     ) -> Result<(), EbpfError<E>> {
-        let fat_ptr: DynTraitFatPointer = unsafe { std::mem::transmute(&*syscall_context_object) };
         let syscall_registry = self.executable.get_syscall_registry();
-        let slot = match hash {
-            Some(hash) => {
-                syscall_registry
-                    .lookup_syscall(hash)
-                    .ok_or(EbpfError::SyscallNotRegistered(hash as usize))?
-                    .context_object_slot
+
+        for (_hash, syscall) in syscall_registry.entries.iter() {
+            let syscall_object_init_fn: SyscallInit<C, E> =
+                unsafe { std::mem::transmute(syscall.init) };
+            let syscall_context_object: Box<dyn SyscallObject<E> + 'a> =
+                syscall_object_init_fn(syscall_context.clone());
+            let fat_ptr: DynTraitFatPointer =
+                unsafe { std::mem::transmute(&*syscall_context_object) };
+
+            let slot = match hash {
+                Some(hash) => {
+                    syscall_registry
+                        .lookup_syscall(hash)
+                        .ok_or(EbpfError::SyscallNotRegistered(hash as usize))?
+                        .context_object_slot
+                }
+                None => syscall_registry
+                    .lookup_context_object_slot(fat_ptr.vtable.methods[0] as u64)
+                    .ok_or(EbpfError::SyscallNotRegistered(
+                        fat_ptr.vtable.methods[0] as usize,
+                    ))?,
+            };
+            if !self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + slot].is_null() {
+                return Err(EbpfError::SyscallAlreadyBound(slot));
+            } else {
+                self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + slot] = fat_ptr.data;
+                // Keep the dyn trait objects so that they can be dropped properly later
+                self.syscall_context_object_pool
+                    .push(syscall_context_object);
             }
-            None => syscall_registry
-                .lookup_context_object_slot(fat_ptr.vtable.methods[0] as u64)
-                .ok_or(EbpfError::SyscallNotRegistered(
-                    fat_ptr.vtable.methods[0] as usize,
-                ))?,
-        };
-        if !self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + slot].is_null() {
-            Err(EbpfError::SyscallAlreadyBound(slot))
-        } else {
-            self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + slot] = fat_ptr.data;
-            // Keep the dyn trait objects so that they can be dropped properly later
-            self.syscall_context_object_pool
-                .push(syscall_context_object);
-            Ok(())
         }
+
+        Ok(())
     }
 
     /// Lookup a syscall context object by its function pointer. Used for testing and validation.
@@ -631,9 +677,11 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// ];
     ///
     /// // Instantiate a VM.
+    /// let config = Config::default();
     /// let mut bpf_functions = std::collections::BTreeMap::new();
-    /// register_bpf_function(&mut bpf_functions, 0, "entrypoint", false).unwrap();
-    /// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
+    /// let syscall_registry = SyscallRegistry::default();
+    /// register_bpf_function(&config, &mut bpf_functions, &syscall_registry, 0, "entrypoint").unwrap();
+    /// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, config, syscall_registry, bpf_functions).unwrap();
     /// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(&executable, &mut [], mem).unwrap();
     ///
     /// // Provide a reference to the packet data.
@@ -662,7 +710,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
         const U32MAX: u64 = u32::MAX as u64;
 
         // R1 points to beginning of input memory, R10 to the stack of the first frame
-        let mut reg: [u64; 11] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, self.stack.get_stack_top()];
+        let mut reg: [u64; 11] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, self.stack.get_frame_ptr()];
 
         if self.memory_mapping.map::<UserError>(AccessType::Store, ebpf::MM_INPUT_START, 1).is_ok() {
             reg[1] = ebpf::MM_INPUT_START;
@@ -671,6 +719,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
         // Check config outside of the instruction loop
         let instruction_meter_enabled = self.executable.get_config().enable_instruction_meter;
         let instruction_tracing_enabled = self.executable.get_config().enable_instruction_tracing;
+        let dynamic_stack_frames = self.executable.get_config().dynamic_stack_frames;
 
         // Loop on instructions
         let entry = self.executable.get_entrypoint_instruction_offset()?;
@@ -695,7 +744,18 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                 self.tracer.trace(state);
             }
 
+
             match insn.opc {
+                _ if dst == STACK_PTR_REG && dynamic_stack_frames => {
+                    match insn.opc {
+                        ebpf::SUB64_IMM => self.stack.resize_stack(-insn.imm),
+                        ebpf::ADD64_IMM => self.stack.resize_stack(insn.imm),
+                        _ => {
+                            #[cfg(debug_assertions)]
+                            unreachable!("unexpected insn on r11")
+                        }
+                    }
+                }
 
                 // BPF_LD class
                 // Since this pointer is constant, and since we already know it (ebpf::MM_INPUT_START), do not
@@ -826,7 +886,22 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                     if reg[src] as u32 == 0 {
                         return Err(EbpfError::DivideByZero(pc + ebpf::ELF_INSN_DUMP_OFFSET));
                     }
-                                    reg[dst] = (reg[dst] as u32 / reg[src] as u32)               as u64;
+                    reg[dst] = (reg[dst] as u32 / reg[src] as u32) as u64;
+                },
+                ebpf::SDIV32_IMM  => {
+                    if reg[dst] as i32 == i32::MIN && insn.imm == -1 {
+                        return Err(EbpfError::DivideOverflow(pc + ebpf::ELF_INSN_DUMP_OFFSET));
+                    }
+                    reg[dst] = (reg[dst] as i32 / insn.imm as i32) as u64;
+                }
+                ebpf::SDIV32_REG  => {
+                    if reg[src] as i32 == 0 {
+                        return Err(EbpfError::DivideByZero(pc + ebpf::ELF_INSN_DUMP_OFFSET));
+                    }
+                    if reg[dst] as i32 == i32::MIN && reg[src] as i32 == -1 {
+                        return Err(EbpfError::DivideOverflow(pc + ebpf::ELF_INSN_DUMP_OFFSET));
+                    }
+                    reg[dst] = (reg[dst] as i32 / reg[src] as i32) as u64;
                 },
                 ebpf::OR32_IMM   =>   reg[dst] = (reg[dst] as u32             | insn.imm as u32) as u64,
                 ebpf::OR32_REG   =>   reg[dst] = (reg[dst] as u32             | reg[src] as u32) as u64,
@@ -885,6 +960,22 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                     }
                                     reg[dst] /= reg[src];
                 },
+                ebpf::SDIV64_IMM  => {
+                    if reg[dst] as i64 == i64::MIN && insn.imm == -1 {
+                        return Err(EbpfError::DivideOverflow(pc + ebpf::ELF_INSN_DUMP_OFFSET));
+                    }
+
+                    reg[dst] = (reg[dst] as i64 / insn.imm) as u64
+                }
+                ebpf::SDIV64_REG  => {
+                    if reg[src] == 0 {
+                        return Err(EbpfError::DivideByZero(pc + ebpf::ELF_INSN_DUMP_OFFSET));
+                    }
+                    if reg[dst] as i64 == i64::MIN && reg[src] as i64 == -1 {
+                        return Err(EbpfError::DivideOverflow(pc + ebpf::ELF_INSN_DUMP_OFFSET));
+                    }
+                    reg[dst] = (reg[dst] as i64 / reg[src] as i64) as u64;
+                },
                 ebpf::OR64_IMM   => reg[dst] |=  insn.imm as u64,
                 ebpf::OR64_REG   => reg[dst] |=  reg[src],
                 ebpf::AND64_IMM  => reg[dst] &=  insn.imm as u64,
@@ -935,7 +1026,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
 
                 ebpf::CALL_REG   => {
                     let target_address = reg[insn.imm as usize];
-                    reg[ebpf::STACK_REG] =
+                    reg[ebpf::FRAME_PTR_REG] =
                         self.stack.push(&reg[ebpf::FIRST_SCRATCH_REG..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS], next_pc)?;
                     if target_address < self.program_vm_addr {
                         return Err(EbpfError::CallOutsideTextSegment(pc + ebpf::ELF_INSN_DUMP_OFFSET, target_address / ebpf::INSN_SIZE as u64 * ebpf::INSN_SIZE as u64));
@@ -969,11 +1060,8 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                         }
                     } else if let Some(target_pc) = self.executable.lookup_bpf_function(insn.imm as u32) {
                         // make BPF to BPF call
-                        reg[ebpf::STACK_REG] = self.stack.push(
-                            &reg[ebpf::FIRST_SCRATCH_REG
-                                ..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS],
-                            next_pc,
-                        )?;
+                        reg[ebpf::FRAME_PTR_REG] =
+                            self.stack.push(&reg[ebpf::FIRST_SCRATCH_REG..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS], next_pc)?;
                         next_pc = self.check_pc(pc, target_pc)?;
                     } else if self.executable.get_config().disable_unresolved_symbols_at_runtime {
                         return Err(EbpfError::UnsupportedInstruction(pc + ebpf::ELF_INSN_DUMP_OFFSET));
@@ -984,12 +1072,12 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
 
                 ebpf::EXIT => {
                     match self.stack.pop::<E>() {
-                        Ok((saved_reg, stack_ptr, ptr)) => {
+                        Ok((saved_reg, frame_ptr, ptr)) => {
                             // Return from BPF to BPF call
                             reg[ebpf::FIRST_SCRATCH_REG
                                 ..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS]
                                 .copy_from_slice(&saved_reg);
-                            reg[ebpf::STACK_REG] = stack_ptr;
+                            reg[ebpf::FRAME_PTR_REG] = frame_ptr;
                             next_pc = self.check_pc(pc, ptr)?;
                         }
                         _ => {
@@ -1004,6 +1092,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                 }
                 _ => return Err(EbpfError::UnsupportedInstruction(pc + ebpf::ELF_INSN_DUMP_OFFSET)),
             }
+
             if instruction_meter_enabled && self.last_insn_count >= remaining_insn_count {
                 // Use `pc + instruction_width` instead of `next_pc` here because jumps and calls don't continue at the end of this instruction
                 return Err(EbpfError::ExceededMaxInstructions(pc + instruction_width + ebpf::ELF_INSN_DUMP_OFFSET, initial_insn_count));
